// PV112 2017, lesson 5 - procedural textures, framebuffer objects

#include <iostream>
#include "PV112.h"

#define _USE_MATH_DEFINES
#include <math.h>

// Include GLEW, use static library
#define GLEW_STATIC
#include <GL/glew.h>
#pragma comment(lib, "glew32s.lib") // Link with GLEW library
#pragma comment(lib, "DevIL.lib") // Link with DevIL library

// Include FreeGLUT
#include <GL/freeglut.h>

// Include the most important GLM functions
#include <glm/gtc/type_ptr.hpp>
#include <glm/gtc/matrix_transform.hpp>

using namespace std;
using namespace PV112;

// Current window size
int win_width = 480;
int win_height = 480;

// Shader program and its uniforms
GLuint program;

GLint model_matrix_loc;
GLint PVM_matrix_loc;
GLint normal_matrix_loc;

GLint material_ambient_color_loc;
GLint material_diffuse_color_loc;
GLint material_specular_color_loc;
GLint material_shininess_loc;

GLint light_position_loc;
GLint light_diffuse_color_loc;
GLint light_ambient_color_loc;
GLint light_specular_color_loc;

GLint eye_position_loc;

// Simple geometries that we will use in this lecture
PV112Geometry my_cube;
PV112Geometry my_teapot;

// Simple camera that allows us to look at the object from different views
PV112Camera my_camera;

// Current time of the application in seconds, for animations
float app_time_s = 0.0f;

// TASK 5 declarations. Screen quad is just a square.
/*
// Screen quad
const float screen_quad_data[8] = {
    -1.0f, 1.0f,
    -1.0f, -1.0f,
    1.0f, 1.0f,
    1.0f, -1.0f,
};
GLuint screen_quad_vbo;
GLuint screen_quad_vao;

// Framebuffer object
int fbo_width, fbo_height;
GLuint fbo_color_texture;
GLuint fbo_depth_stencil_texture;
GLuint fbo;

// Postprocess shader
GLuint postprocess_program;
GLuint sceneTex_loc;
GLuint app_time_s_loc;
*/

// Called when the user presses a key
void key_pressed(unsigned char key, int mouseX, int mouseY)
{
    switch (key)
    {
    case 27: // Escape
        exit(0);
        break;
    case 'l':
        // TASK 8: This suddenly doesn't work right! Try to fix it.
        glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
        glutPostRedisplay();
        break;
    case 'f':
        glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
        glutPostRedisplay();
        break;
    case 't':
        glutFullScreenToggle();
        break;
    }
}

// Called when the user presses a mouse button
void mouse_button_changed(int button, int state, int x, int y)
{
    my_camera.OnMouseButtonChanged(button, state, x, y);
}

// Called when the user moves with the mouse (when some mouse button is pressed)
void mouse_moved(int x, int y)
{
    my_camera.OnMouseMoved(x, y);
}

// TASK 5: Uncomment this function and use it appropriately.
/*
// Updates the size of FBO textures using the window size
void resize_fbo_textures()
{
    fbo_width = win_width;
    fbo_height = win_height;
    glBindTexture(GL_TEXTURE_2D, fbo_color_texture);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, fbo_width, fbo_height, 0, GL_RGBA,
        GL_UNSIGNED_BYTE, nullptr);
    glBindTexture(GL_TEXTURE_2D, fbo_depth_stencil_texture);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_DEPTH24_STENCIL8, fbo_width, fbo_height,
        0, GL_DEPTH_STENCIL, GL_UNSIGNED_INT_24_8, nullptr);
    glBindTexture(GL_TEXTURE_2D, 0);
}
*/

// Initializes OpenGL stuff
void init()
{
    glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
    glClearDepth(1.0);

    // Create shader program
    program = CreateAndLinkProgram("vertex.glsl", "fragment.glsl");
    if (0 == program)
        WaitForEnterAndExit();

    // Get attribute and uniform locations
    int position_loc = glGetAttribLocation(program, "position");
    int normal_loc = glGetAttribLocation(program, "normal");
    int tex_coord_loc = glGetAttribLocation(program, "tex_coord");

    model_matrix_loc = glGetUniformLocation(program, "model_matrix");
    PVM_matrix_loc = glGetUniformLocation(program, "PVM_matrix");
    normal_matrix_loc = glGetUniformLocation(program, "normal_matrix");

    material_ambient_color_loc = glGetUniformLocation(program, "material_ambient_color");
    material_diffuse_color_loc = glGetUniformLocation(program, "material_diffuse_color");
    material_specular_color_loc = glGetUniformLocation(program, "material_specular_color");
    material_shininess_loc = glGetUniformLocation(program, "material_shininess");

    light_position_loc = glGetUniformLocation(program, "light_position");
    light_ambient_color_loc = glGetUniformLocation(program, "light_ambient_color");
    light_diffuse_color_loc = glGetUniformLocation(program, "light_diffuse_color");
    light_specular_color_loc = glGetUniformLocation(program, "light_specular_color");

    eye_position_loc = glGetUniformLocation(program, "eye_position");

    // Create geometries
    my_cube = CreateCube(position_loc, normal_loc, tex_coord_loc);
    my_teapot = CreateTeapot(position_loc, normal_loc, tex_coord_loc);

    //----------------------
    //--  Postprocessing  --

    // TASK 5: Initialize framebuffer and set up postprocessing
    /*
    // Create postprocessing shader program
    postprocess_program = CreateAndLinkProgram("postprocess_vertex.glsl", "postprocess_fragment.glsl");
    if (0 == postprocess_program)
        WaitForEnterAndExit();

    // Get attribute and uniform locations
    position_loc = glGetAttribLocation(postprocess_program, "position");
    sceneTex_loc = glGetUniformLocation(postprocess_program, "sceneTex");
    app_time_s_loc = glGetUniformLocation(postprocess_program, "app_time_s");

    // Set the sceneTex texture unit to 0
    glUseProgram(postprocess_program);
    glUniform1i(sceneTex_loc, 0);
    glUseProgram(0);

    // Prepare screen quad geometry and VAO
    glGenBuffers(1, &screen_quad_vbo);
    glBindBuffer(GL_ARRAY_BUFFER, screen_quad_vbo);
    glBufferData(GL_ARRAY_BUFFER, sizeof(float) * 8, screen_quad_data, GL_STATIC_DRAW);
    glGenVertexArrays(1, &screen_quad_vao);
    glBindVertexArray(screen_quad_vao);
    glEnableVertexAttribArray(position_loc);
    glVertexAttribPointer(position_loc, 2, GL_FLOAT, GL_FALSE, 0, nullptr);
    glBindVertexArray(0);
    glBindBuffer(GL_ARRAY_BUFFER, 0);

    // Prepare FBO textures
    glGenTextures(1, &fbo_color_texture);
    glGenTextures(1, &fbo_depth_stencil_texture);
    resize_fbo_textures();

    // Set parameters of the color texture. We do not use the depth texture for
    // rendering, so we do not set its parameters.
    glBindTexture(GL_TEXTURE_2D, fbo_color_texture);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glBindTexture(GL_TEXTURE_2D, 0);

    // Prepare FBO
    glGenFramebuffers(1, &fbo);
    glBindFramebuffer(GL_FRAMEBUFFER, fbo);
    glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0,
        GL_TEXTURE_2D, fbo_color_texture, 0);
    glFramebufferTexture2D(GL_FRAMEBUFFER, GL_DEPTH_STENCIL_ATTACHMENT,
        GL_TEXTURE_2D, fbo_depth_stencil_texture, 0);
    GLenum bufs[1] = {GL_COLOR_ATTACHMENT0};
    glDrawBuffers(1, bufs);
    GLenum status = glCheckFramebufferStatus(GL_FRAMEBUFFER);
    if (status != GL_FRAMEBUFFER_COMPLETE)
    {
        cout << "Failed to create a complete framebuffer" << endl;
        WaitForEnterAndExit();
    }
    glBindFramebuffer(GL_FRAMEBUFFER, 0);
    */
}

// Called when the window needs to be rerendered
void render()
{
    // TASK 5: Render the scene into framebuffer texture. It will need
    //   several minor changes in the existing render code.
    //glBindFramebuffer(GL_FRAMEBUFFER, fbo);
    glViewport(0, 0, win_width, win_height);

    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glEnable(GL_DEPTH_TEST);

    // Some variables
    glm::mat4 projection_matrix, view_matrix, model_matrix, PVM_matrix;
    glm::mat3 normal_matrix;

    projection_matrix = glm::perspective(glm::radians(45.0f),
            float(win_width) / float(win_height), 0.1f, 100.0f);
    view_matrix = glm::lookAt(my_camera.GetEyePosition(),
            glm::vec3(0.0f, 0.0f, 0.0f), glm::vec3(0.0f, 1.0f, 0.0f));

    // Light position, with a simple animation
    glm::vec4 light_pos =
        glm::rotate(glm::mat4(1.0f), app_time_s, glm::vec3(0.0f, 1.0f, 0.0f)) *
        glm::translate(glm::mat4(1.0f), glm::vec3(5.0f, 2.0f, 0.0f)) *
        glm::vec4(0.0f, 0.0f, 0.0f, 1.0f);

    glUseProgram(program);

    glUniform3fv(eye_position_loc, 1, glm::value_ptr(my_camera.GetEyePosition()));

    glUniform4fv(light_position_loc, 1, glm::value_ptr(light_pos));
    glUniform3f(light_ambient_color_loc, 0.3f, 0.3f, 0.3f);
    glUniform3f(light_diffuse_color_loc, 1.0f, 1.0f, 1.0f);
    glUniform3f(light_specular_color_loc, 1.0f, 1.0f, 1.0f);

    // Render the scene object

    model_matrix = glm::mat4(1.0f);

    glUniform3f(material_ambient_color_loc, 1.0f, 1.0f, 1.0f);
    glUniform3f(material_diffuse_color_loc, 1.0f, 1.0f, 1.0f);
    glUniform3f(material_specular_color_loc, 1.0f, 1.0f, 1.0f);
    glUniform1f(material_shininess_loc, 40.0f);

    PVM_matrix = projection_matrix * view_matrix * model_matrix;
    normal_matrix = glm::inverse(glm::transpose(glm::mat3(model_matrix)));
    glUniformMatrix4fv(model_matrix_loc, 1, GL_FALSE, glm::value_ptr(model_matrix));
    glUniformMatrix4fv(PVM_matrix_loc, 1, GL_FALSE, glm::value_ptr(PVM_matrix));
    glUniformMatrix3fv(normal_matrix_loc, 1, GL_FALSE, glm::value_ptr(normal_matrix));

    glBindVertexArray(my_cube.VAO);
    DrawGeometry(my_cube);

    glBindVertexArray(0);
    glUseProgram(0);

    // TASK 5: Draw the framebuffer using postprocessing program
    /*
    glBindFramebuffer(GL_FRAMEBUFFER, 0);
    glViewport(0, 0, win_width, win_height);
    glClear(GL_COLOR_BUFFER_BIT);
    glDisable(GL_DEPTH_TEST);

    glUseProgram(postprocess_program);
    glUniform1f(app_time_s_loc, app_time_s);

    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, fbo_color_texture);

    glBindVertexArray(screen_quad_vao);
    glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
    glBindVertexArray(0);
    glUseProgram(0);
    */

    // Swaps the front and back buffer (double-buffering)
    glutSwapBuffers();
}

// Called when the window changes its size
void reshape(int width, int height)
{
    win_width = width;
    win_height = height;
    //resize_fbo_textures();
}

// Callback function to be called when we make an error in OpenGL
void GLAPIENTRY simple_debug_callback(GLenum source, GLenum type, GLuint id,
        GLenum severity, GLsizei length, const char* message, const void* userParam)
{
    switch (type)            // Report only some errors
    {
    case GL_DEBUG_TYPE_ERROR:
    case GL_DEBUG_TYPE_DEPRECATED_BEHAVIOR:
    case GL_DEBUG_TYPE_UNDEFINED_BEHAVIOR:
        cout << message << endl;        // Put the breakpoint here
        return;
    default:
        return;
    }
}

// Simple timer function for animations
void timer(int)
{
    // Add 20 ms to the application time, register the timer to be called again
    // in 20 ms, and rerender the scene
    app_time_s += 0.020f;
    glutTimerFunc(20, timer, 0);
    glutPostRedisplay();
}

int main(int argc, char ** argv)
{
    // Initialize GLUT
    glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_DEPTH | GLUT_DOUBLE | GLUT_RGBA);
    glutSetOption(GLUT_ACTION_ON_WINDOW_CLOSE, GLUT_ACTION_GLUTMAINLOOP_RETURNS);

    // Set OpenGL Context parameters
    glutInitContextVersion(3, 3);              // Use OpenGL 3.3
    glutInitContextProfile(GLUT_CORE_PROFILE); // Use OpenGL core profile
    glutInitContextFlags(GLUT_DEBUG);          // Use OpenGL debug context

    // Initialize window
    glutInitWindowSize(win_width, win_height);
    glutCreateWindow("PV112 - cv5");

    // Initialize GLEW
    glewExperimental = GL_TRUE;
    glewInit();

    // Initialize DevIL library
    InitImageLibrary();

    // Set the debug callback
    SetDebugCallback(simple_debug_callback);

    // Initialize out OpenGL stuff
    init();

    // Register callbacks
    glutDisplayFunc(render);
    glutReshapeFunc(reshape);
    glutKeyboardFunc(key_pressed);
    glutMouseFunc(mouse_button_changed);
    glutMotionFunc(mouse_moved);
    glutTimerFunc(20, timer, 0);

    // Run the main loop
    glutMainLoop();

    return 0;
}
